#include <libxslt/extensions.h>
#include <libxml/xpathInternals.h>
#include <libxml/uri.h>
#include <libxslt/xsltutils.h>
#include <SWFFile.h>
#include <stdlib.h>
#include <string.h>
#include "swft.h"

#define TMP_STRLEN 0xFF
#define SWFT_MAPSIZE 64
#define SWFT_NAMESPACE ((const xmlChar*)"http://subsignal.org/swfml/swft")

using namespace SWF;

struct swft_mapentry {
	int o;
	int n;
};

struct swft_ctx {
	int last_id;
	
	int map_n, map_alloc;
	swft_mapentry *map;
};

void swft_resizemap( swft_ctx *ctx );
static void swft_nextid( xmlXPathParserContextPtr ctx, int nargs );
static void swft_resetmap( xsltTransformContextPtr ctx, xmlNodePtr node, xmlNodePtr inst, xsltElemPreCompPtr comp );
static void swft_mapid( xmlXPathParserContextPtr ctx, int nargs );
static void swft_document( xmlXPathParserContextPtr ctx, int nargs );
void *swft_init( xsltTransformContextPtr ctx, const xmlChar *URI );
void swft_shutdown( xsltTransformContextPtr ctx, const xmlChar *URI, void *data );

void swft_register() {
	xsltRegisterExtModule((const xmlChar *)SWFT_NAMESPACE,
		swft_init, swft_shutdown );
}

void swft_resizemap( swft_ctx *ctx ) {
	if( ctx->map_n >= ctx->map_alloc ) {
		int ns = ctx->map_alloc + SWFT_MAPSIZE;
		swft_mapentry *nmap = new swft_mapentry[ ns ];
		memset( nmap, 0, ns*sizeof(swft_mapentry) );
		
		if( ctx->map_n ) {
			memcpy( nmap, ctx->map, sizeof(swft_mapentry)*ctx->map_n );
			delete ctx->map;
		}
		
		ctx->map = nmap;
		ctx->map_alloc = ns;
	}
}

static void swft_nextid( xmlXPathParserContextPtr ctx, int nargs ) {
	char tmp[TMP_STRLEN];
	swft_ctx *c = (swft_ctx*)xsltGetExtData( xsltXPathGetTransformContext(ctx), SWFT_NAMESPACE );
	snprintf(tmp,TMP_STRLEN,"%i", c->last_id++ );
	valuePush(ctx, xmlXPathNewString((const xmlChar *)tmp));
}

static void swft_resetmap( xsltTransformContextPtr ctx, xmlNodePtr node, xmlNodePtr inst, xsltElemPreCompPtr comp ) {
	swft_ctx *c = (swft_ctx*)xsltGetExtData( ctx, SWFT_NAMESPACE );
	if( c->map ) delete c->map;
	c->map = NULL;
	c->map_n = c->map_alloc = 0;
}

static void swft_mapid( xmlXPathParserContextPtr ctx, int nargs ) {
	char tmp[TMP_STRLEN];
	xmlXPathObjectPtr obj;
	int oldID;
	int newID;
	swft_ctx *c = (swft_ctx*)xsltGetExtData( xsltXPathGetTransformContext(ctx), SWFT_NAMESPACE );

	xmlXPathStringFunction(ctx, 1);
	if (ctx->value->type != XPATH_STRING) {
		xsltTransformError(xsltXPathGetTransformContext(ctx), NULL, NULL,
			 "swft:mapid() : invalid arg expecting a number (the old id)\n");
		ctx->error = XPATH_INVALID_TYPE;
		return;
	}
	obj = valuePop(ctx);
	if (obj->stringval == NULL) {
		valuePush(ctx, xmlXPathNewNodeSet(NULL));
		return;
	}
	
	oldID = (int)xmlXPathStringEvalNumber( obj->stringval );
	newID = -1;
	// try to find oldID in our map
	for( int i=0; newID==-1 && i<c->map_n; i++ ) {
		if( c->map[i].o == oldID ) {
			newID = c->map[i].n;
		}
	}
	
	if( newID == -1 ) {
		// assign next id and store in map.
		newID = c->last_id++;
		swft_resizemap(c);
		c->map[c->map_n].o = oldID;
		c->map[c->map_n].n = newID;
		c->map_n++;
	}
	
//	fprintf(stderr, "------------ swft: mapping %i->%i\n", oldID, newID );
	
	snprintf(tmp,TMP_STRLEN,"%i", newID );
	valuePush(ctx, xmlXPathNewString((const xmlChar *)tmp));
}

static void swft_document( xmlXPathParserContextPtr ctx, int nargs ) {
	xsltTransformContextPtr tctx;
	xmlChar *uri;
	xsltDocumentPtr xsltdoc;
	xmlDocPtr doc = NULL;
	xmlXPathObjectPtr obj;
	xmlChar *base = NULL;
	char sig;
	File import;

	xmlXPathStringFunction(ctx, 1);
	if (ctx->value->type != XPATH_STRING) {
		xsltTransformError(xsltXPathGetTransformContext(ctx), NULL, NULL,
			 "swft:document() : invalid arg expecting a string\n");
		ctx->error = XPATH_INVALID_TYPE;
		return;
	}
	obj = valuePop(ctx);
	if (obj->stringval == NULL) {
		valuePush(ctx, xmlXPathNewNodeSet(NULL));
		return;
	}
		
	tctx = xsltXPathGetTransformContext(ctx);
	
	uri = obj->stringval;
		
	FILE *fp = fopen( (const char *)uri, "rb" );
	if( !fp ) {
		xsltTransformError(xsltXPathGetTransformContext(ctx), NULL, NULL,
				   "swft:document() : failed to read file '%s'\n", (const char *)uri);
		valuePush(ctx, xmlXPathNewNodeSet(NULL));
		return;
	}
	
	import.load( fp );
	doc = import.getXML();
	
	if( !doc ) {
		xsltTransformError(xsltXPathGetTransformContext(ctx), NULL, NULL,
				   "swft:document() : could not parse SWF '%s'\n", (const char *)uri);
		valuePush(ctx, xmlXPathNewNodeSet(NULL));
		return;
	}
	valuePush( ctx, xmlXPathNewNodeSet( (xmlNodePtr)doc ) );
	return;
}

void *swft_init( xsltTransformContextPtr ctx, const xmlChar *URI ) {
	xsltRegisterExtFunction( ctx, (const xmlChar *) "next-id", SWFT_NAMESPACE, swft_nextid);
	xsltRegisterExtFunction( ctx, (const xmlChar *) "map-id", SWFT_NAMESPACE, swft_mapid);
	xsltRegisterExtFunction( ctx, (const xmlChar *) "document", SWFT_NAMESPACE, swft_document);
	xsltRegisterExtElement(  ctx, (const xmlChar *) "reset-map", SWFT_NAMESPACE, swft_resetmap);
	
	swft_ctx *c = new swft_ctx;
	c->last_id = 1;
	c->map_n = c->map_alloc = 0;
	c->map = NULL;
	return c;
}
void swft_shutdown( xsltTransformContextPtr ctx, const xmlChar *URI, void *data ) {
	swft_ctx *c = (swft_ctx*)data;
	delete c;
}

